package org.grails.plugins.wordpress

import grails.orm.HibernateCriteriaBuilder
import groovy.net.xmlrpc.XMLRPCServerProxy
import org.codehaus.groovy.grails.commons.ConfigurationHolder
import org.hibernate.Session
import org.hibernate.criterion.Restrictions
import org.springframework.transaction.support.TransactionCallback
import org.springframework.transaction.support.TransactionTemplate
import org.springframework.orm.hibernate3.SessionFactoryUtils

import org.grails.plugins.wordpress.model.Category
import org.grails.plugins.wordpress.model.Comment
import org.grails.plugins.wordpress.model.Page
import org.grails.plugins.wordpress.model.Post
import org.grails.plugins.wordpress.model.Tag
import org.grails.plugins.wordpress.model.User

class WordpressService {

    static transactional = false

    def wordpressTransactionManager
    def wordpressSessionFactory
    def cacheService

    String url = ConfigurationHolder.config.wordpress.url
    String blogId = ConfigurationHolder.config.wordpress.blogId
    String username = ConfigurationHolder.config.wordpress.username
    String password = ConfigurationHolder.config.wordpress.password

    def withTransaction(Closure callable) {
        new TransactionTemplate(wordpressTransactionManager).execute({status ->
                callable.call(status)
            } as TransactionCallback)
    }

    public Tag getTagByName(String tagName) {
        withTransaction {
            new HibernateCriteriaBuilder(Tag.class, wordpressSessionFactory).get {
                eq('name', tagName)
            }
        }
    }

    public Category getCategoryByName(String categoryName) {
        withTransaction {
            new HibernateCriteriaBuilder(Category.class, wordpressSessionFactory).get {
                eq('name', categoryName)
            }
        }
    }

    private Category updateCategory(Map map) {
        log.trace "Wordpress Category : ${map}"
        def session = SessionFactoryUtils.getSession(wordpressSessionFactory, false)
        Category category = session.get(Category.class, map.categoryId as Long) ?: new Category(id:map.categoryId as Long)
        category.name = map.categoryName
        category.parent = session.get(Category.class, map.parentId as Long)
        category.htmlUrl = map.htmlUrl
        category.categoryDescription = map.categoryDescription
        category.rssUrl = map.rssUrl
        session.save(category)
        category
    }

    private void updateCategoryParent(Map map) {
        if (map.parentId) {
            def session = SessionFactoryUtils.getSession(wordpressSessionFactory, false)
            Category category = session.get(Category.class, map.categoryId as Long)
            Category parent = session.get(Category.class, map.parentId as Long)
            category.parent = parent
            session.save(category)
        }
    }

    private Comment updateComment(Map map) {
        log.trace "Wordpress Comment : ${map}"
        def session = SessionFactoryUtils.getSession(wordpressSessionFactory, false)
        Comment comment = session.get(Comment.class, map.comment_id as Long) ?: new Comment(id:map.comment_id as Long)
        comment.dateCreatedGmt = map.date_created_gmt
        comment.user = session.get(User.class, map.user_id as Long)
        comment.parent = session.get(Comment.class, map.parent as Long)
        comment.status = map.status
        comment.content = map.content
        comment.link = map.link
        comment.post = session.get(Post.class, map.post_id as Long)
        comment.author = map.author
        comment.authorUrl = map.author_url
        comment.authorEmail = map.author_email
        comment.authorIp = map.author_ip
        comment.type = map.type
        session.save(comment)
        comment
    }

    private Page updatePage(Map map) {
        log.trace "Wordpress Page : ${map}"
        def session = SessionFactoryUtils.getSession(wordpressSessionFactory, false)
        Page page = session.get(Page.class, map.page_id as Long) ?: new Page(id:map.page_id as Long)
        page.author = session.get(User.class, map.wp_author_id as Long)
        page.description = map.description
        page.title = map.title
        page.link = map.link
        page.permaLink = map.permaLink
        page.excerpt = map.excerpt
        page.textMore = map.text_more
        page.allowComments = map.mt_allow_comments
        page.allowPings = map.mt_allow_pings
        page.slug = map.wp_slug
        page.password = map.wp_password
        page.dateCreated = map.dateCreated
        page.dateCreatedGmt = map.date_created_gmt
        page.status = map.page_status
        page.customFields = map.custom_fields
        page.pageOrder = map.wp_page_order
        page.pageTemplate = map.wp_page_template
        page.parent = session.get(Page.class, map.wp_page_parent_id as Long)
        session.save(page)
        page
    }

    private void updatePageParent(Map map) {
        if (map.wp_page_parent_id) {
            def session = SessionFactoryUtils.getSession(wordpressSessionFactory, false)
            Page page = session.get(Page.class, map.page_id as Long)
            Page parent = session.get(Page.class, map.wp_page_parent_id as Long)
            page.parent = parent
            session.save(page)
        }
    }

    private Post updatePost(Map map) {
        log.trace "Wordpress Post : ${map}"
        def session = SessionFactoryUtils.getSession(wordpressSessionFactory, false)
        Post post = session.get(Post.class, map.postid as Long) ?: new Post(id:map.postid as Long)
        if (post.comments == null) {
            post.comments = []
        }
        post.categories = []
        post.tags = []
        post.author = session.get(User.class, map.wp_author_id as Long)
        post.description = map.description
        post.title = map.title
        post.link = map.link
        post.permaLink = map.permaLink
        post.excerpt = map.mt_excerpt
        post.textMore = map.mt_text_more
        post.allowComments = map.mt_allow_comments
        post.allowPings = map.mt_allow_pings
        post.wpSlug = map.wp_slug
        post.wpPassword = map.wp_password
        post.dateCreated = map.dateCreated
        post.dateCreatedGmt = map.date_created_gmt
        post.status = map.post_status
        post.customFields = map.custom_fields
        map.categories.each {
            Category category = getCategoryByName(it)
            if (category) {
                post.categories << category
            } else {
                log.error "Could not retrieve Category : ${it}"
            }
        }
        map.mt_keywords.split(',').each {
            def tagName = it.trim()
            if (tagName) {
                Tag tag = getTagByName(tagName)
                if (tag) {
                    post.tags << tag
                } else {
                    log.error "Could not retrieve Tag : ${tagName}"
                }
            }
        }
        session.save(post)
        post
    }

    private Tag updateTag(Map map) {
        log.trace "Wordpress Tag : ${map}"
        def session = SessionFactoryUtils.getSession(wordpressSessionFactory, false)
        Tag tag = session.get(Tag.class, map.tag_id as Long) ?: new Tag(id:map.tag_id as Long)
        tag.name = map.name
        tag.count = map.count
        tag.slug = map.slug
        tag.htmlUrl = map.htmlUrl
        tag.rssUrl = map.rssUrl
        session.save(tag)
        tag
    }

    private User updateUser(Map map) {
        log.trace "Wordpress User : ${map}"
        def session = SessionFactoryUtils.getSession(wordpressSessionFactory, false)
        User user = session.get(User.class, map.user_id as Long) ?: new User(id:map.user_id as Long)
        user.login = map.user_login
        user.name = map.display_name
        session.save(user)
        user
    }

    private void deleteMissing(Class clazz, List existingIds) {
        if (existingIds) {
            def session = SessionFactoryUtils.getSession(wordpressSessionFactory, false)
            def criteria = session.createCriteria(clazz)
            criteria.add(Restrictions.not(Restrictions.in('id', existingIds)))
            criteria.list().each { session.delete(it) }
        }
    }

    void updateCache() {
        withTransaction {
            try {
                log.info("Starting wordpress cache update")
                
                def serverProxy = new XMLRPCServerProxy(url, true)
                def wpUsers = serverProxy."wp.getAuthors"(blogId, username, password)
                def wpCategories = serverProxy."wp.getCategories"(blogId, username, password)
                def wpTags = serverProxy."wp.getTags"(blogId, username, password)
                def wpPageList = serverProxy."wp.getPageList"(blogId, username, password)
                def wpPages = serverProxy."wp.getPages"(blogId, username, password, wpPageList.size())
                def mostRecentPost = serverProxy."metaWeblog.getRecentPosts"(blogId, username, password, 1)
                def wpPosts = mostRecentPost ? serverProxy."metaWeblog.getRecentPosts"(blogId, username, password, mostRecentPost[0].postid) : []

                // Take care of missing items first
                deleteMissing(Post, wpPosts.collect { map -> map.postid as Long })
                deleteMissing(Page, wpPages.collect { map -> map.page_id as Long })
                deleteMissing(Tag, wpTags.collect { map -> map.tag_id as Long })
                deleteMissing(Category, wpCategories.collect { map -> map.categoryId as Long })
                deleteMissing(User, wpUsers.collect { map -> map.user_id as Long })

                // Update existing items
                wpUsers.each {
                    updateUser(it)
                }
                wpCategories.each {
                    updateCategory(it)
                }
                wpCategories.each {
                    updateCategoryParent(it)
                }
                wpTags.each {
                    updateTag(it)
                }
                wpPages.each {
                    updatePage(it)
                }
                wpPages.each {
                    updatePageParent(it)
                }
                wpPosts.each {
                    updatePost(it)
                }

//                getRecentPosts('', [max:wpPosts.size()]).each { Post post ->
//                    def commentCount = serverProxy."wp.getCommentCount"(blogId, username, password, post.id)
//                    def totalComments = commentCount.'total_comments'
//                    log.debug "Update Cache : Post Id ${post.id} : Total Comments : ${totalComments} : Current Comments : ${post.comments.size()}"
//                    if (totalComments > post.comments.size()) {
//                        serverProxy."wp.getComments"(blogId, username, password, [offset:post.comments.size(), max:post.comments.size() - totalComments]).each { wpComment ->
//                            updateComment(wpComment)
//                        }
//                    }
//                }

            log.info "Wordpress cache updated : ${wpPosts.size()} posts, ${wpPages.size()} pages"
            } catch (Exception e) {
                throw new RuntimeException("An error occurred while updating the wordpress cache", e)
            }
        }
    }

    def getCategories(Map attrs = [:]) {
        withTransaction { status ->
            if (attrs.nocache) {
                def serverProxy = new XMLRPCServerProxy(url, true)
                try {
                    serverProxy."wp.getCategories"(blogId, username, password).each { wpCategory ->
                        updateCategory(wpCategory)
                    }
                } catch (Exception e) {
                    throw new RuntimeException("Unable to update categories", e)
                }
            }
            new HibernateCriteriaBuilder(Category.class, wordpressSessionFactory).list {
                order('name', 'asc')
            }
        }
    }

    // Comment status can be 'hold', 'approve', 'spam'
    def getComments(Map attrs = [:]) {
        withTransaction {
            if (attrs.nocache) {
                def serverProxy = new XMLRPCServerProxy(url, true)
                try {
                    serverProxy."wp.getComments"(blogId, username, password, attrs).each { wpComment ->
                        updateComment(wpComment)
                    }
                } catch (Exception e) {
                    throw new RuntimeException("Unable to update comments", e)
                }
            }
            new HibernateCriteriaBuilder(Comment.class, wordpressSessionFactory).list {
                if (attrs.status) {
                    eq('status', attrs.status)
                }
                if (attrs.postId) {
                    post {
                        idEq(attrs.postId as Long)
                    }
                }
                if (attrs.offset) {
                    firstResult(attrs.offset as Integer)
                }
                maxResults(attrs.max as Integer ?: 100)
            }
        }
    }

    def getPage(Long id, Map attrs = [:]) {
        withTransaction {
            if (attrs.nocache) {
                def serverProxy = new XMLRPCServerProxy(url, true)
                try {
                    // Get all parents too
                    def wpPages = []
                    def map
                    for (Long tmpId = id;tmpId;) {
                        map = serverProxy."wp.getPage"(blogId, tmpId, username, password)
                        wpPages << map
                        tmpId = map.wp_page_parent_id as Long
                    }
                    wpPages.reverse().each {
                        updatePage(it)
                    }
                } catch (Exception e) {
                    throw new RuntimeException("Unable to update page", e)
                }
            }
            new HibernateCriteriaBuilder(Page.class, wordpressSessionFactory).get {
                eq('id', id)
            }
        }
    }

    def getPages(Map attrs = [:]) {
        withTransaction {
            new HibernateCriteriaBuilder(Page.class, wordpressSessionFactory).list {
                eq('status', attrs.status ?: 'publish')
                if (attrs.postId) {
                   idEq('post', attrs.postId as Long)
                }
                if (attrs.offset) {
                    firstResult(attrs.offset as Integer)
                }
                maxResults(attrs.max as Integer ?: 100)
            }
        }
    }

    def getPost(Long id, Map attrs = [:]) {
        withTransaction {
            if (attrs.nocache) {
                def serverProxy = new XMLRPCServerProxy(url, true)
                try {
                    updatePost(serverProxy."metaWeblog.getPost"(id, username, password))
                } catch (Exception e) {
                    throw new RuntimeException("Unable to update post", e)
                }
            }
            new HibernateCriteriaBuilder(Post.class, wordpressSessionFactory).get {
                eq('id', id)
            }
        }
    }

    def getRecentPosts(Map attrs = [:]) {
        withTransaction {
            if (attrs.nocache) {
                def serverProxy = new XMLRPCServerProxy(url, true)
                try {
                    def mostRecentPost = serverProxy."metaWeblog.getRecentPosts"(blogId, username, password, 1)
                    serverProxy."metaWeblog.getRecentPosts"(blogId, username, password, mostRecentPost[0].postid).each { post ->
                        log.trace "Wordpress Post : ${post}"
                        updatePost(post)
                    }
                } catch (Exception e) {
                    throw new RuntimeException("Unable to update recent posts", e)
                }
            }
            new HibernateCriteriaBuilder(Post.class, wordpressSessionFactory).list {
                eq('status', attrs.status ?: 'publish')
                if (attrs.postId) {
                   idEq('post', attrs.postId as Long)
                }
                if (attrs.offset) {
                    firstResult(attrs.offset as Integer)
                }
                maxResults(attrs.max as Integer ?: 100)
                order('dateCreated', 'desc')
            }
        }
    }

    def getRecentPostsWithTagInCategory(String tagName, String categoryName, Map attrs = [:]) {
        def posts = []
        withTransaction {
            Category category = getCategoryByName(categoryName)
            Tag tag = getTagByName(tagName)
            if (tag && category) {
                posts = new HibernateCriteriaBuilder(Post.class, wordpressSessionFactory).list {
                    eq('status', attrs.status ?: 'publish')
                    if (attrs.postId) {
                       idEq('post', attrs.postId as Long)
                    }
                    if (attrs.offset) {
                        firstResult(attrs.offset as Integer)
                    }
                    categories {
                        idEq(category.id)
                    }
                    tags {
                        idEq(tag.id)
                    }
                    maxResults(attrs.max as Integer ?: 100)
                    order('dateCreated', 'desc')
                }
            } else {
                if (!category) {
                    log.warn "Category ${categoryName} not found."
                }
                if (!tag) {
                    log.warn "Tag ${tagName} not found."
                }
            }
        }
        return posts
    }

    def getPostCountInCategory(String categoryName, Map attrs = [:]) {
        withTransaction {
            new HibernateCriteriaBuilder(Post.class, wordpressSessionFactory).get {
                eq('status', attrs.status ?: 'publish')
                categories {
                    eq('name', categoryName)
                }
                projections {
                    rowCount()
                }
            }
        }
    }

    def getRecentPostsInCategory(String categoryName, Map attrs = [:]) {
        withTransaction {
            Category category = getCategoryByName(categoryName)
            if (category) {
                new HibernateCriteriaBuilder(Post.class, wordpressSessionFactory).list {
                    eq('status', attrs.status ?: 'publish')
                    if (attrs.postId) {
                       idEq('post', attrs.postId as Long)
                    }
                    if (attrs.offset) {
                        firstResult(attrs.offset as Integer)
                    }
                    categories {
                        idEq(category.id)
                    }
                    maxResults(attrs.max as Integer ?: 100)
                    order('dateCreated', 'desc')
                }
            } else {
                log.warn "Category ${categoryName} not found."
                []
            }
        }
    }

    def getPostCountWithTag(String tagName, Map attrs = [:]) {
        withTransaction {
            new HibernateCriteriaBuilder(Post.class, wordpressSessionFactory).get {
                eq('status', attrs.status ?: 'publish')
                tags {
                    eq('name', tagName)
                }
                projections {
                    rowCount()
                }
            }
        }
    }

    def getRecentPostsWithTag(String tagName, Map attrs = [:]) {
        withTransaction {
            Tag tag = getTagByName(tagName)
            if (tag) {
                new HibernateCriteriaBuilder(Post.class, wordpressSessionFactory).list {
                    eq('status', attrs.status ?: 'publish')
                    if (attrs.postId) {
                       idEq('post', attrs.postId as Long)
                    }
                    if (attrs.offset) {
                        firstResult(attrs.offset as Integer)
                    }
                    tags {
                        idEq(tag.id)
                    }
                    maxResults(attrs.max as Integer ?: 100)
                    order('dateCreated', 'desc')
                }
            } else {
                log.warn "Tag ${tagName} not found."
                []
            }
        }
    }

    def getPostCount(Map attrs = [:]) {
        withTransaction {
            new HibernateCriteriaBuilder(Post.class, wordpressSessionFactory).get {
                eq('status', attrs.status ?: 'publish')
                projections {
                    rowCount()
                }
            }
        }
    }

    def getTags(Map attrs = [:]) {
        withTransaction {
            if (attrs.nocache) {
                def serverProxy = new XMLRPCServerProxy(url, true)
                try {
                    serverProxy."wp.getTags"(blogId, username, password).each { wpTag ->
                        updateTag(wpTag)
                    }
                } catch (Exception e) {
                    throw new RuntimeException("Unable to update tags", e)
                }
            }
            new HibernateCriteriaBuilder(Tag.class, wordpressSessionFactory).list {
                order('name', 'asc')
            }
        }
    }

    def getUsers(Map attrs = [:]) {
        withTransaction {
            if (attrs.nocache) {
                def serverProxy = new XMLRPCServerProxy(url, true)
                try {
                    serverProxy."wp.getAuthors"(blogId, username, password).each { wpUser ->
                        updateUser(wpUser)
                    }
                } catch (Exception e) {
                    throw new RuntimeException("Unable to update users", e)
                }
            }
            new HibernateCriteriaBuilder(User.class, wordpressSessionFactory).list {
                order('name', 'asc')
            }
        }
    }
}
